@@ -1,14 +1,14 @@
 import Cookie
-from datetime import datetime, timedelta
+import base64
 import hmac
 import md5
 import os
+import cPickle
 import random
-import re
 import sys
 import time
 import UserDict
-import warnings
+from datetime import datetime, timedelta
 
 try:
     from paste.registry import StackedObjectProxy
@@ -16,7 +16,32 @@ try:
 except:
     beaker_session = None
 
+# Determine if PyCrypto is available
+crypto_ok = False
+strong_hash = None
+try:
+    from Crypto.Cipher import AES
+    from Crypto.Hash import SHA, SHA256
+    import Crypto
+    if Crypto.__version__ >= '2.0.1':
+        crypto_ok = True
+        strong_hash = SHA.new
+    from beaker.crypto import generateCryptoKeys
+    from beaker.crypto.CTRCipher import CTRCipher
+    from beaker.crypto.PBKDF2 import strxor
+except:
+    pass
+
+# Attempt to use a stronger hash function if we're in Python 2.5
+if not strong_hash:
+    try:
+        import hashlib
+        strong_hash = hashlib.sha1
+    except:
+        pass
+
 from beaker.container import namespace_registry
+from beaker.exceptions import BeakerException
 from beaker.util import coerce_session_params
 
 __all__ = ['SignedCookie', 'Session']
@@ -28,16 +53,25 @@ class SignedCookie(Cookie.BaseCookie):
         Cookie.BaseCookie.__init__(self, input)
     
     def value_decode(self, val):
-        sig = val[0:32]
-        value = val[32:]
-    
-        if hmac.new(self.secret, value).hexdigest() != sig:
-            return None, val
-        
-        return val[32:], val
+        val = val.strip('"')
+        if strong_hash:
+            sig = val[0:40]
+            value = val[40:]
+            if strong_hash(self.secret + value).hexdigest() != sig:
+                return None, val
+            return val[40:], val
+        else:
+            sig = val[0:32]
+            value = val[32:]
+            if hmac.new(self.secret, value).hexdigest() != sig:
+                return None, val
+            return val[32:], val
     
     def value_encode(self, val):
-        return val, ("%s%s" % (hmac.new(self.secret, val).hexdigest(), val))
+        if strong_hash:
+            return str(val), ("%s%s" % (strong_hash(self.secret + val).hexdigest(), val))
+        else:
+            return str(val), ("%s%s" % (hmac.new(self.secret, val).hexdigest(), val))
 
                 
 class Session(UserDict.DictMixin):
@@ -46,7 +80,8 @@ class Session(UserDict.DictMixin):
     def __init__(self, request, id=None, invalidate_corrupt=False, 
                  use_cookies=True, type=None, data_dir=None, 
                  key='beaker.session.id', timeout=None, cookie_expires=True,
-                 secret=None, log_file=None, namespace_class=None, **kwargs):
+                 cookie_domain=None, secret=None, log_file=None, 
+                 namespace_class=None, **kwargs):
         if type is None:
             if data_dir is None:
                 self.type = 'memory'
@@ -68,6 +103,7 @@ class Session(UserDict.DictMixin):
         self.timeout = timeout
         self.use_cookies = use_cookies
         self.cookie_expires = cookie_expires
+        self.cookie_domain = cookie_domain
         self.log_file = log_file
         self.was_invalidated = False
         self.secret = secret
@@ -114,6 +150,8 @@ class Session(UserDict.DictMixin):
         self.is_new = True
         if self.use_cookies:
             self.cookie[self.key] = self.id
+            if self.cookie_domain:
+                self.cookie[self.key]['domain'] = self.cookie_domain
             self.cookie[self.key]['path'] = '/'
             if self.cookie_expires is not True:
                 if self.cookie_expires is False:
@@ -136,11 +174,7 @@ class Session(UserDict.DictMixin):
         """deletes the persistent storage for this session, but remains valid. """
         self.namespace.acquire_write_lock()
         try:
-            for k in self.namespace.keys():
-                if not re.match(r'_creation_time|_accessed_time', k):
-                    del self.namespace[k]
-                    
-            self.namespace['_accessed_time'] = time.time()
+            self.namespace.remove()
         finally:
             self.namespace.release_write_lock()
     
@@ -266,6 +300,127 @@ class Session(UserDict.DictMixin):
         if self.log_file is not None:
             self.log_file.write(message)
 
+
+class CookieSession(Session):
+    """Pure cookie-based session using RC4 encryption
+    
+    Options recognized when using cookie-based sessions are slightly
+    more restricted than general sessions.
+    
+    ``key``
+        The name the cookie should be set to.
+    ``timeout``
+        How long session data is considered valid. This is used 
+        regardless of the cookie being present or not to determine
+        whether session data is still valid.
+    ``encrypt_key``
+        The key to use for the session encryption.
+    ``validate_key``
+        The key used to sign the encrypted session
+    ``cookie_domain``
+        Domain to use for the cookie.
+    
+    .. note ::
+        For RC4 encryption security, key's should not exceed 54 
+        characters in lenght. The key used for encryption is specified
+        as the session secret.
+    
+    """
+    def __init__(self, request, key='beaker.session.id', timeout=None,
+                 cookie_expires=True, cookie_domain=None, encrypt_key=None,
+                 validate_key=None, **kwargs):
+        if not crypto_ok:
+            raise BeakerException("PyCrypto is not installed, can't use cookie-only Session.")
+        
+        self.request = request
+        self.key = key
+        self.timeout = timeout
+        self.cookie_expires = cookie_expires
+        self.cookie_domain = cookie_domain
+        self.encrypt_key = encrypt_key
+        self.validate_key = validate_key
+        self.request['set_cookie'] = False
+        
+        try:
+            cookieheader = request['cookie']
+        except KeyError:
+            cookieheader = ''
+        
+        if encrypt_key is None:
+            raise BeakerException("No encrypt_key specified for Cookie only Session.")
+        if validate_key is None:
+            raise BeakerException("No validate_key specified for Cookie only Session.")
+        
+        try:
+            self.cookie = SignedCookie(validate_key, input=cookieheader)
+        except Cookie.CookieError:
+            self.cookie = SignedCookie(validate_key, input=None)
+        
+        self.dict = {}
+        self.is_new = True
+        
+        # If we have a cookie, load it
+        if self.key in self.cookie and self.cookie[self.key].value is not None:
+            self.is_new = False
+            try:
+                self.dict = self._decrypt_data()
+            except:
+                self.dict = {}
+            if self.timeout is not None and time.time() - self.dict['_accessed_time'] > self.timeout:
+                self.dict = {}
+            self._create_cookie()
+    
+    created = property(lambda self: self.dict['_creation_time'])
+    
+    def _encrypt_data(self):
+        """Cerealize, encipher, and base64 the session dict"""
+        encrypt_key, nonce = generateCryptoKeys(self.encrypt_key, self.validate_key, 1)
+        nonce = base64.b64encode(nonce)[:8]
+        ctrcipher = CTRCipher(encrypt_key, nonce, AES)
+        data = cPickle.dumps(self.dict, protocol=2)
+        return nonce + base64.b64encode(ctrcipher.encrypt(data))
+    
+    def _decrypt_data(self):
+        """Bas64, decipher, then un-cerealize the data for the session dict"""
+        nonce = self.cookie[self.key].value[:8]
+        encrypt_key, salt = generateCryptoKeys(self.encrypt_key, self.validate_key, 1)
+        ctrcipher = CTRCipher(encrypt_key, nonce, AES)
+        payload = base64.b64decode(self.cookie[self.key].value[8:])
+        data = ctrcipher.decrypt(payload)
+        return cPickle.loads(data.strip('\0'))
+    
+    def save(self):
+        "saves the data for this session to persistent storage"
+        self._create_cookie()
+    
+    def _create_cookie(self):
+        if '_creation_time' not in self.dict:
+            self.dict['_creation_time'] = time.time()
+        self.dict['_accessed_time'] = time.time()
+        val = self._encrypt_data()
+        if len(val) > 4064:
+            raise BeakerException("Cookie value is too long to store")
+        
+        self.cookie[self.key] = val
+        if self.cookie_domain:
+            self.cookie[self.key]['domain'] = self.cookie_domain
+        self.cookie[self.key]['path'] = '/'
+        if self.cookie_expires is not True:
+            if self.cookie_expires is False:
+                expires = datetime.fromtimestamp( 0x7FFFFFFF )
+            elif isinstance(self.cookie_expires, timedelta):
+                expires = datetime.today() + self.cookie_expires
+            elif isinstance(self.cookie_expires, datetime):
+                expires = self.cookie_expires
+            else:
+                raise ValueError("Invalid argument for cookie_expires: %s"
+                                 % repr(self.cookie_expires))
+            self.cookie[self.key]['expires'] = \
+                expires.strftime("%a, %d-%b-%Y %H:%M:%S GMT" )
+        self.request['cookie_out'] = self.cookie[self.key].output(header='')
+        self.request['set_cookie'] = True
+
+
 class SessionObject(object):
     """Session proxy/lazy creator
     
@@ -288,7 +443,10 @@ class SessionObject(object):
             environ = self.__dict__['_environ']
             self.__dict__['_headers'] = req = {'cookie_out':None}
             req['cookie'] = environ.get('HTTP_COOKIE')
-            self.__dict__['_sess'] = Session(req, use_cookies=True, **params)
+            if params.get('type') == 'cookie':
+                self.__dict__['_sess'] = CookieSession(req, **params)
+            else:
+                self.__dict__['_sess'] = Session(req, use_cookies=True, **params)
         return self.__dict__['_sess']
     
     def __getattr__(self, attr):
@@ -326,95 +484,3 @@ class SessionObject(object):
             session.namespace.remove()
             return None
         return session
-
-
-class SessionMiddleware(object):
-    deprecated = True
-    session = beaker_session
-    
-    def __init__(self, wrap_app, config=None, environ_key='beaker.session', **kwargs):
-        """Initialize the Session Middleware
-        
-        The Session middleware will make a lazy session instance available 
-        every request under the ``environ['beaker.cache']`` key by default. The location in
-        environ can be changed by setting ``environ_key``.
-        
-        ``config``
-            dict  All settings should be prefixed by 'cache.'. This method of
-            passing variables is intended for Paste and other setups that
-            accumulate multiple component settings in a single dictionary. If
-            config contains *no cache. prefixed args*, then *all* of the config
-            options will be used to intialize the Cache objects.
-        
-        ``environ_key``
-            Location where the Session instance will keyed in the WSGI environ
-        
-        ``**kwargs``
-            All keyword arguments are assumed to be cache settings and will
-            override any settings found in ``config``
-        """
-        if self.deprecated:
-            warnings.warn('SessionMiddleware is moving to beaker.middleware in '
-              '0.8', DeprecationWarning, 2)
-        
-        config = config or {}
-        
-        # Load up the default params
-        self.options = dict(invalidate_corrupt=True, type=None, 
-                           data_dir=None, key='beaker.session.id', 
-                           timeout=None, secret=None, log_file=None)
-
-        # Pull out any config args meant for beaker session. if there are any
-        for dct in [config, kwargs]:
-            for key, val in dct.iteritems():
-                if key.startswith('beaker.session.'):
-                    self.options[key[15:]] = val
-                if key.startswith('session.'):
-                    self.options[key[8:]] = val
-                if key.startswith('session_'):
-                    warnings.warn('Session options should start with session. '
-                                  'instead of session_.', DeprecationWarning, 2)
-                    self.options[key[8:]] = val
-        
-        # Coerce and validate session params
-        coerce_session_params(self.options)
-        
-        # Assume all keys are intended for cache if none are prefixed with 'cache.'
-        if not self.options and config:
-            self.options = config
-        
-        self.options.update(kwargs)
-        self.wrap_app = wrap_app
-        self.environ_key = environ_key
-        
-    def __call__(self, environ, start_response):
-        session = SessionObject(environ, **self.options)
-        if environ.get('paste.registry'):
-            environ['paste.registry'].register(self.session, session)
-        environ[self.environ_key] = session
-        environ['beaker.get_session'] = self._get_session
-        
-        def session_start_response(status, headers, exc_info = None):
-            if session.__dict__['_sess'] is not None:
-                if session.__dict__['_headers']['set_cookie']:
-                    cookie = session.__dict__['_headers']['cookie_out']
-                    if cookie:
-                        headers.append(('Set-cookie', cookie))
-            return start_response(status, headers, exc_info)
-        try:
-            response = self.wrap_app(environ, session_start_response)
-        except:
-            ty, val = sys.exc_info()[:2]
-            if isinstance(ty, str):
-                raise ty, val, sys.exc_info()[2]
-            if ty.__name__ == 'HTTPFound' and \
-                    session.__dict__['_sess'] is not None:
-                cookie = session.__dict__['_headers']['cookie_out']
-                if cookie:
-                    val.headers.append(('Set-cookie', cookie))
-            raise ty, val, sys.exc_info()[2]
-        else:
-            return response
-    
-    def _get_session(self):
-        return Session({}, use_cookies=False, **self.options)
